"use strict";

exports.__esModule = true;
exports.TABLE_NAME = exports.AsyncRemoteModule = void 0;
exports.checkRemoteModuleWebpack = checkRemoteModuleWebpack;
exports["default"] = void 0;
exports.existModule = existModule;
exports.globalModule = exports.globalDB = exports.globalCached = void 0;
exports.isAbsoluteUrl = isAbsoluteUrl;
exports.joinUrl = joinUrl;
exports.requireJs = requireJs;
exports.resolveModuleUrl = resolveModuleUrl;
require("core-js/modules/es6.object.assign.js");
require("core-js/modules/es7.array.includes.js");
require("core-js/modules/es6.object.to-string.js");
require("core-js/modules/es6.promise.js");
require("core-js/modules/es6.array.reduce.js");
require("core-js/modules/es6.regexp.split.js");
require("core-js/modules/es6.date.now.js");
require("core-js/modules/es6.array.index-of.js");
require("core-js/modules/es6.array.for-each.js");
require("core-js/modules/es6.object.keys.js");
require("core-js/modules/es6.string.starts-with.js");
var _objdp = require("./_objdp");
exports.objectDefineProperty = _objdp.objectDefineProperty;
var DEFAULT_HEAD_TIMEOUT = 30000;
var TABLE_NAME = 'fetched';
exports.TABLE_NAME = TABLE_NAME;
function checkRemoteModuleWebpack(context) {
  if (context === void 0) {
    context = window;
  }
  var globalModule = context.__remoteModuleWebpack__;
  if (!globalModule) {
    context.__remoteModuleWebpack__ = globalModule = {
      __moduleManifests__: {},
      cached: {}
    };
  }
  if (!globalModule.shareModules) globalModule.shareModules = {};
  return globalModule;
}
var globalModule = checkRemoteModuleWebpack();
exports.globalModule = globalModule;
var globalCached = window.__remoteModuleWebpack__.cached;
exports.globalCached = globalCached;
var globalDB = function globalDB() {
  return window.__remoteModuleWebpack__.db;
};
exports.globalDB = globalDB;
var queue = [];
function pushQueue(url, resolve, reject) {
  var item = {
    url: url
  };
  queue.push(item);
  var walk = function walk() {
    while (queue.length && queue[0].done) {
      queue.shift().done();
    }
  };
  return {
    success: function success(r) {
      item.done = function () {
        return resolve(r);
      };
      walk();
    },
    fail: function fail(r) {
      item.done = function () {
        return reject(r);
      };
      walk();
    }
  };
}
function innumerable(obj, key, value, options) {
  if (options === void 0) {
    options = {
      configurable: true
    };
  }
  (0, _objdp.objectDefineProperty)(obj, key, Object.assign({
    value: value
  }, options));
  return obj;
}

/** @type {import('../types/fetch').default} */
function fetch(url, _temp) {
  var _ref = _temp === void 0 ? {} : _temp,
    _ref$timeout = _ref.timeout,
    timeout = _ref$timeout === void 0 ? 120000 : _ref$timeout,
    sync = _ref.sync,
    cacheDB = _ref.cacheDB,
    nocache = _ref.nocache,
    _ref$method = _ref.method,
    method = _ref$method === void 0 ? 'GET' : _ref$method,
    headers = _ref.headers;
  if (!globalCached._fetched) innumerable(globalCached, '_fetched', {});
  var fetched = globalCached._fetched;
  var next = function next(url) {
    if (fetched[url]) return fetched[url];
    var isHeadRequest = ['HEAD', 'OPTIONS'].includes(method);
    var prom = new Promise(function (resolve, reject) {
      var res = pushQueue(url, resolve, function (r) {
        delete fetched[url];
        return reject(r);
      });
      var xhr = new XMLHttpRequest();
      var timerId;
      var isTimedOut = false;
      xhr.onreadystatechange = function () {
        if (xhr.readyState === 4) {
          timerId && clearTimeout(timerId) || (timerId = 0);
          if (xhr.status === 0) {
            var cb = function cb() {
              // timeout
              var err = new Error("fetch [" + url + "] " + (isTimedOut ? 'timed out' : 'failed') + ".");
              err.xhr = xhr;
              err.url = url;
              res.fail(err);
            };
            if ((cacheDB || globalModule.cacheDB) && globalDB()) {
              globalDB().get(TABLE_NAME, url).then(function (v) {
                console.error("warning:[import-remtoe]fetch [ " + url + " ] failed, use cacheDB's cache instead.");
                res.success(v.text);
              })["catch"](cb);
            } else cb();
          } else if (xhr.status === 404) {
            // no update available
            var err = new Error("fetch [" + url + "] not found.");
            err.xhr = xhr;
            err.url = url;
            res.fail(err);
          } else if (xhr.status !== 200 && xhr.status !== 304) {
            // other failure
            var _err = new Error("fetch [" + url + "] failed:" + xhr.status + ".");
            _err.xhr = xhr;
            _err.url = url;
            res.fail(_err);
          } else {
            // success
            if (isHeadRequest) {
              var _headers = xhr.getAllResponseHeaders().split('\n').reduce(function (p, v) {
                var _v$split = v.split(': '),
                  key = _v$split[0],
                  value = _v$split[1];
                if (!key) return p;
                p[key] = value;
                return p;
              }, {});
              res.success(_headers);
            } else {
              if ((cacheDB || globalModule.cacheDB) && globalDB()) {
                globalDB().put(TABLE_NAME, {
                  url: url,
                  text: xhr.responseText,
                  timestamp: Date.now(),
                  modified: xhr.getResponseHeader('last-modified')
                });
              }
              res.success(xhr.responseText);
            }
          }
        }
      };
      try {
        if (!sync) xhr.timeout = timeout;
        xhr.open(method, nocache ? url + ((~url.indexOf('?') ? '&' : '?') + "_=" + Date.now()) : url, !sync);
        xhr.setRequestHeader('Content-Type', 'text/plain;charset=UTF-8');
        if (headers) Object.keys(headers).forEach(function (key) {
          return xhr.setRequestHeader(key, headers[key]);
        });

        // if (nocache) {
        //   xhr.setRequestHeader('If-Modified-Since', '0');
        //   xhr.setRequestHeader('Cache-Control', 'no-cache');
        // }
        xhr.send(null);
        timerId = setTimeout(function () {
          isTimedOut = true;
          xhr.abort();
          xhr.onreadystatechange = null;
          timerId = 0;
          res.fail({
            type: 'timeout',
            target: xhr
          });
        }, timeout);
      } catch (e) {
        res.fail(e);
      }
    });
    if (!isHeadRequest) fetched[url] = prom;
    return prom;
  };
  if (url.then) return url.then(next);
  return next(url);
}
fetch.queue = queue;

/** @type {import('../types/fetch').requireJs} */
function requireJs(url, options) {
  if (options === void 0) {
    options = {};
  }
  var cached = options.cached || globalCached;
  if (!cached._rs) innumerable(cached, '_rs', {});
  var next = function next(url) {
    if (cached._rs[url]) return cached._rs[url];
    return cached._rs[url] = fetch(url, options).then(function (src) {
      return new Promise(function (resolve, reject) {
        // eslint-disable-next-line no-new-func
        var fn = new Function('module', 'exports', 'require', src);
        var _module = {
          exports: {}
        };
        try {
          fn(_module, _module.exports, options.require || function (name) {
            throw new Error("[import-remote:requireJs]module [" + name + "] cannot be found!");
          });
          resolve(_module.exports);
        } catch (ex) {
          reject(ex);
        }
      });
    })["catch"](function (ex) {
      delete cached._rs[url];
      throw ex;
    });
  };
  if (url.then) return url.then(next);
  return next(url);
}

/**
 * @param {string} url
 * @returns
 */
function isAbsoluteUrl(url) {
  return typeof url === 'string' && /^(((https?:)?\/\/)|(data:))/.test(url);
}

/**
 * @param {string} host
 * @param {string} [path]
 */
function joinUrl(host, path) {
  if (path && /^["'].+["']$/.test(path)) path = path.substr(1, path.length - 2);
  if (!host || isAbsoluteUrl(path) /* || /^\//.test(path) */) return path;
  if (/^\/[A-Za-z]/.test(host) && path.startsWith(host)) return path;
  if (/\/$/.test(host)) host = host.substr(0, host.length - 1);
  if (/^\.\//.test(path)) path = path.substr(1, path.length);
  return "" + host + (/^\//.test(path) ? path : "/" + path);
}

/**
 * @param {string} host
 * @param {string} [moduleName]
 * @param {boolean} [sync]
 */
function resolveModuleUrl(host, moduleName) {
  if (moduleName === void 0) {
    moduleName = 'index.js';
  }
  if (!/\.js$/.test(moduleName)) moduleName += '.js';
  return joinUrl(host, moduleName);
}
function existModule(host, moduleName, options) {
  if (moduleName === void 0) {
    moduleName = 'index.js';
  }
  if (options === void 0) {
    options = {};
  }
  return new Promise(function (resolve) {
    return fetch(resolveModuleUrl(host, moduleName), Object.assign({
      timeout: DEFAULT_HEAD_TIMEOUT,
      nocache: true,
      method: 'HEAD'
    }, options)).then(function (r) {
      return resolve(r);
    })["catch"](function () {
      return resolve(null);
    });
  });
}

/**
 * @type {new (
 *   libraryUrl: string,
 *   host: string,
 *   options?: import('../types/fetch').RemoteModuleOptions
 * ) => import('../types/fetch').AsyncRemoteModule}
 */
var AsyncRemoteModule = /*#__PURE__*/function () {
  function AsyncRemoteModule(libraryUrl, host, options) {
    if (options === void 0) {
      options = {};
    }
    if (!libraryUrl) throw new Error('[AsyncRemoteModule]libraryUrl can not be null!');
    this.libraryUrl = libraryUrl;
    this.host = host;
    this.options = options || {};
  }
  var _proto = AsyncRemoteModule.prototype;
  _proto.readyRuntime = function readyRuntime() {
    var _this = this;
    return requireJs(this.libraryUrl).then(function (_ref2) {
      var RemoteModule = _ref2.RemoteModule;
      if (_this.runtime) return _this.runtime;
      return _this.runtime = new RemoteModule(_this.host, _this.options);
    });
  };
  _proto.exist = function exist(moduleName, options) {
    if (moduleName === void 0) {
      moduleName = 'index.js';
    }
    if (options === void 0) {
      options = {};
    }
    return existModule(this.host, moduleName, options);
  };
  return AsyncRemoteModule;
}();
exports.AsyncRemoteModule = AsyncRemoteModule;
['requireMeta', 'requireEntries', 'require', 'import'].forEach(function (key) {
  return AsyncRemoteModule.prototype[key] = function () {
    var _arguments = arguments;
    return this.readyRuntime().then(function (runtime) {
      return runtime[key].apply(runtime, _arguments);
    });
  };
});
var _default = fetch;
exports["default"] = _default;