"use strict";

require("core-js/modules/es6.weak-map.js");
require("core-js/modules/es6.string.iterator.js");
require("core-js/modules/es6.array.iterator.js");
require("core-js/modules/web.dom.iterable.js");
require("core-js/modules/es6.object.define-property.js");
require("core-js/modules/es6.object.get-own-property-descriptor.js");
exports.__esModule = true;
exports["default"] = exports.ATTR_CSS_TRANSFORMED = void 0;
exports.transformStyleHost = transformStyleHost;
require("core-js/modules/es6.regexp.replace.js");
require("core-js/modules/es6.object.to-string.js");
require("core-js/modules/es6.promise.js");
var _fetch = _interopRequireWildcard(require("./fetch"));
var _utils = require("./utils");
function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function _getRequireWildcardCache(nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }
function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { "default": obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj["default"] = obj; if (cache) { cache.set(obj, newObj); } return newObj; }
var ATTR_CSS_TRANSFORMED = 'data-import-remote-transformed';
exports.ATTR_CSS_TRANSFORMED = ATTR_CSS_TRANSFORMED;
function hasFetched(href, head, scopeName) {
  if (scopeName === void 0) {
    scopeName = '';
  }
  if (!head) head = document.documentElement.getElementsByTagName('head')[0];
  var existingLinkTags = head.getElementsByTagName('link');
  for (var i = 0; i < existingLinkTags.length; i++) {
    var tag = existingLinkTags[i];
    var dataHref = tag.getAttribute('data-href') || tag.getAttribute('href');
    var dataScopeName = tag.getAttribute(_utils.ATTR_SCOPE_NAME) || '';
    if (tag.rel === 'stylesheet' && dataHref === href && dataScopeName === scopeName) return true;
  }
  var existingStyleTags = head.getElementsByTagName('style');
  for (var _i = 0; _i < existingStyleTags.length; _i++) {
    var _tag = existingStyleTags[_i];
    var _dataHref = _tag.getAttribute('data-href');
    var _dataScopeName = _tag.getAttribute(_utils.ATTR_SCOPE_NAME) || '';
    if (_dataHref === href && _dataScopeName === scopeName) return true;
  }
  return false;
}
function transformStyleHost(source, host) {
  if (!host || !source) return source;
  if (/\/$/.test(host)) host = host.substr(0, host.length - 1);
  return source.replace(/url\(([^)]+)\)/ig, function (m, p1) {
    return "url(" + (0, _utils.joinUrl)(host, p1) + ")";
  });
  // .replace(/@import\s+(["'])([^"']+)["']/ig, (m, p0, p1) => `@import ${p0 + joinUrl(host, p1) + p0}`);
}

/** @type {import('../types/importCss').default}  */
function fetchStyle(url, _temp) {
  var _ref = _temp === void 0 ? {} : _temp,
    _ref$cached = _ref.cached,
    cached = _ref$cached === void 0 ? _fetch.globalCached : _ref$cached,
    _ref$timeout = _ref.timeout,
    timeout = _ref$timeout === void 0 ? _utils.DEFAULT_TIMEOUT : _ref$timeout,
    sync = _ref.sync,
    cacheDB = _ref.cacheDB,
    head = _ref.head,
    nocache = _ref.nocache,
    scopeName = _ref.scopeName,
    host = _ref.host,
    beforeSource = _ref.beforeSource,
    method = _ref.method,
    devtool = _ref.devtool,
    sourcemapHost = _ref.sourcemapHost,
    publicPath = _ref.publicPath;
  if (!cached._css) (0, _utils.innumerable)(cached, '_css', {});
  var next = function next(url) {
    var cacheUrl = (0, _utils.getCacheUrl)(url, scopeName);
    if (cached._css[cacheUrl]) return cached._css[cacheUrl];
    return cached._css[cacheUrl] = new Promise(function (resolve, reject) {
      // const resolve = r => {
      //   delete cached._css[cacheUrl];
      //   return _resolve(r);
      // };
      // const reject = r => {
      //   delete cached._css[cacheUrl];
      //   return _reject(r);
      // };
      if (!head) head = document.getElementsByTagName('head')[0];
      if (hasFetched(url, head, scopeName)) return resolve();
      (0, _fetch["default"])(url, {
        timeout: timeout,
        sync: sync,
        cacheDB: cacheDB,
        nocache: nocache,
        method: method
      }).then(function (source) {
        try {
          source = transformStyleHost(source, host);
          source = (0, _utils.transformSourcemapUrl)(url, source, {
            devtool: devtool,
            sourcemapHost: sourcemapHost,
            scopeName: scopeName,
            host: host,
            publicPath: publicPath
          });
          if (beforeSource) source = beforeSource(source, 'css', url, {});
          var styleTag = document.createElement('style');
          styleTag.type = 'text/css';
          styleTag.setAttribute('data-href', url);
          styleTag.setAttribute(ATTR_CSS_TRANSFORMED, '');
          if (scopeName) styleTag.setAttribute(_utils.ATTR_SCOPE_NAME, scopeName);
          styleTag.innerHTML = source;
          head.appendChild(styleTag);
          resolve(styleTag);
        } catch (err) {
          console.error(err);
          err.code = 'CSS_CHUNK_LOAD_FAILED';
          reject(err);
        }
      })["catch"](function (ex) {
        delete cached._css[cacheUrl];
        return reject(ex);
      });
    });
  };
  if (url.then) return url.then(next);
  return next(url);
}
/** @type {import('../types/importCss').default}  */
function importCss(href, options) {
  if (options === void 0) {
    options = {};
  }
  return fetchStyle(href, options);
}
var _default = importCss;
exports["default"] = _default;